/*
 * Copyright (c) 1999 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <config.h>

RCSID("$Id: bosserver.c,v 1.9 1999/09/13 16:21:52 lha Exp $");

#include <sys/types.h>
#include <sys/wait.h>

#include <rx/rx.h>
#include <rx/rx_null.h>

#include <ports.h>
#include <ko.h>

#include <ctype.h>

#ifdef KERBEROS
#include <des.h>
#include <krb.h>
#include <rxkad.h>
#endif

#include <netinit.h>
#include <msecurity.h>

#include <service.h>

#include <roken.h>
#include <err.h>
#include <getarg.h>

#include <bos.h>
#include <bos.ss.h>

static char *email = NULL;


typedef struct mtype {
    char *name;
    char *path;
    char *argv; /* XXX */
    pid_t pid;
    struct {
	unsigned long email:1;
	unsigned long savecore:1;
	unsigned long trydebug:1;
    } flags;
    struct mtype *next;
    time_t last_start;
} mtype;

struct mtype *servers = NULL;

static int debug = 1;
static char *bosserverprefix = NULL;

/*
 *
 */

static void
bosdebug (char *fmt, ...)
    __attribute__ ((format (printf, 1, 2)));

static void
bosdebug (char *fmt, ...)
{
    va_list args;
    if (!debug)
	return ;

    va_start (args, fmt);
    vfprintf (stderr, fmt, args);
    va_end(args);
}

/*
 *
 */

char *
get_basename (char *name)
{
    char *ret = strrchr (name, '/');
    if (ret && *ret && *(ret+1)) return ret + 1;
    return name;
}

/*
 * Send mail to irresponsible person
 */

void
sendmail (char *person, char **message)
{
    FILE *f;
    char *cmd;
    int i;

    if (person == NULL) {
	bosdebug ("sendmail: person == NULL");
	return;
    }
    
    asprintf (&cmd, "/usr/sbin/sendmail %s", person);

    f = popen (cmd, "w");
    if (f == NULL) {
	bosdebug ("sendmail: %s", cmd);
	free (cmd);
	return;
    }

    fprintf (f, "From: <bos-watcher>\r\n");
    fprintf (f, "Subject: Death\r\n\r\n");

    i = 0;
    while (message[i]) {
	fprintf (f, "%s\r\n", message[i]);
	i++;
    }
    
    fflush (f);
    pclose (f);

    free (cmd);
}

/*
 *
 */

void
add_server (char *prog, char *flags, char *coredir, char *argv)
{
    struct mtype *server = malloc (sizeof(struct mtype));
    
    if (server == NULL)
	err (1, "add_server: malloc");

    memset (server, 0, sizeof(*server));
    
    server->name  = get_basename(prog);
    server->path  = prog;
    if (strchr (flags, 'e')) server->flags.email = TRUE;
    if (strchr (flags, 'c')) server->flags.savecore = TRUE;
    if (strchr (flags, 'd')) server->flags.trydebug = TRUE;
    server->argv  = argv;


    server->next  = servers;
    servers 	  = server;
}


/*
 * get string from ``p'', the name of the string is ``name''
 * the result is saved in ``res''.
 */

void
get_string (char **p, char *name, char **res)
{
    char *p1 = *p, *p2;

    if (*p1 == '\0')
	errx (1, "no string left for %s", name);
    
    while(isspace((unsigned char)*p1))
	++p1;
    if (*p1 == '\0')
	errx (1, "missing %s", name);
    p2 = p1;
    while(!isspace((unsigned char)*p2))
	++p2;
    if (*p2 == '\0')
	errx (1, "missing %s(2)", name);
    *p2 = '\0';
    *res = estrdup (p1);
    *p = p2 + 1;
}

/*
 *
 */

void
read_config_file (char *filename)
{
    FILE *f = fopen (filename, "r");
    char buf[BUFSIZ];
    char *p1, *p2;
    int count, endp;

    if (f == NULL)
	err (1, "read_config_file: failed to open config file: %s", 
	     filename);


    while (fgets (buf, sizeof(buf), f)) {
	buf[strlen(buf)-1] = '\0';

	if (buf[0] == '#' || buf[0] == '\0')
	    continue;

	count = 0;

	p1 = buf;
	while(isspace((unsigned char)*p1))
	    ++p1;

	if (*p1 == '\0')
	    continue;

	p2 = p1;
	while(!isspace((unsigned char)*p2))
	    ++p2;
	
	endp = *p2 == '\0';
	*p2 = '\0';
	
	if (strcasecmp(p1, "email") == 0) {
	    p1 = p2 + 1;
	    if (endp || p1 == '\0') {
		warnx ("missing argument to email");
		continue;
	    }
	    email = estrdup (p1);
	} else if (strcasecmp(p1, "program") == 0) {
	    char *prog;
	    char *flags;
	    char *coredir;
	    char *argv;

	    p1 = p2 + 1;

	    get_string (&p1, "program", &prog);
	    get_string (&p1, "flags", &flags);
	    get_string (&p1, "coredir", &coredir);
	    get_string (&p1, "argv", &argv); /* XXX */

	    add_server (prog, flags, coredir, argv);

	} else {
	    warnx ("unknown command %s", p1);
	}
    }

    fclose (f);
}


/*
 *
 */

int
BOZO_CreateBnode(struct rx_call *call,
		 const char *type,
		 const char *instance,
		 const char *p1,
		 const char *p2,
		 const char *p3,
		 const char *p4,
		 const char *p5,
		 const char *p6)
{
    bosdebug ("BOZO_CreateNode: %s %s\n", type, instance);

    if (!sec_is_superuser(call))
	return VL_PERM;

    return 0;
}

/*
 *
 */

int
BOZO_DeleteBnode(struct rx_call *call,
		 const char *instance)
{
    bosdebug ("BOZO_DeleteBnode: %s\n", instance);

    return 0;
}


/*
 *
 */

int
BOZO_GetStatus(struct rx_call *call,
	       const char *instance,
	       int32_t *inStat,
	       char *statdescr)
{
    bosdebug ("BOZO_GetStatus: %s\n", instance);

    strcpy (statdescr, "foo");
    *inStat = 0;
    return 0;
}

/*
 *
 */

int
BOZO_EnumerateInstance(struct rx_call *call,
		       const int32_t instance,
		       char *iname)
{
    bosdebug ("BOZO_EnumerateInstance: %d\n", instance);

    return -1;
}

/*
 *
 */

int
BOZO_GetInstanceInfo(struct rx_call *call,
		     const char *instance,
		     char *type,
		     struct bozo_status *status)
{
    bosdebug ("BOZO_GetInstanceInfo: %s\n", instance);

    strcpy (type, "simple");
    memset (status, 0, sizeof(*status));
    return 0;
}


/*
 *
 */

int
BOZO_GetInstanceParm(struct rx_call *call,
		     const char *instance,
		     const int32_t num,
		     char *param)
{
    bosdebug ("BOZO_GetInstanceParm: %s %d\n", instance, num);

    strcpy (param, "foo");
    return 0;
}


/*
 * start the server `server'
 */

static int
start_server (struct mtype *server)
{
    int i;
    time_t newtime = time(NULL);

    bosdebug ("starting server %s (%s)\n", server->name, 
	      server->path);

    if (newtime + 10 < server->last_start 
	&& server->last_start != 0) {
	/* XXX */
	bosdebug ("server %s looping, will try again in 10 min",
		  server->name);
	return 0;
    }
    server->last_start = newtime;
    
    i = fork();
    switch (i) {
    case 0: {
	char *newargv[3];

	newargv[0] = server->path;
	newargv[1] = server->argv;
	newargv[2] = NULL;	

	if (debug == 0) {
	    close (0);
	    close (1);
	    close (2);
	}

	execvp (server->path, newargv);

	bosdebug ("server %s failed with %s (%d)\n", server->path,
		  strerror (errno), errno);
	exit (1);
	break;
    }
    case -1:
	bosdebug ("failed to fork with server %s with reson %s (%d)\n",
		  server->name, strerror (errno), errno);
	return errno;
	break;
    default:
	server->pid = i;
	bosdebug ("started server %s with pid %d\n", 
		  server->name, server->pid);
	break;
    }
    
    return 0;
}


/*
 * GC processes, and start new if appropriate.
 */

static int dead_children = 0;

PROCESS riperprocess;

void
theriper (char *no_arg)
{
    int status;
    char *msg[3] = { NULL, NULL, NULL };
    struct mtype *server = NULL;
    pid_t pid;

    while (1) {
	if (dead_children == 0) {
	    bosdebug ("theriper: sleeping\n");
	    LWP_WaitProcess ((char *)theriper);
	}	

	bosdebug ("theriper: running\n");

	if (dead_children == 0) {
	    bosdebug ("theriper: called without a dead child\n");
	    continue;
	}
	
	if (dead_children < 0)
	    abort();

	dead_children--;
	    
	pid = wait (&status);

	if (WIFEXITED(status)) {
	    asprintf (&msg[1],
		      "normal exit: exit error %d", WEXITSTATUS(status));
	} else if (WIFSIGNALED(status)) {
	    asprintf (&msg[1],
		      "terminated by signal num %d %s", WTERMSIG(status),
		      WCOREDUMP(status) ? "coredumped\n" : "");
	} else if (WIFSTOPPED(status)) {
	    asprintf (&msg[1],
		      "process stoped by signal %d", WSTOPSIG(status));
	} else {
	    asprintf (&msg[1],
		      "by unknown reason");
	}
	bosdebug (msg[1]);

	server = servers;
	while (server) {
	    if (server->pid == pid) {
		msg[0] = server->name;
		if (server->flags.email)
		    sendmail (email, msg);
		bosdebug ("%s was the killed, starting new", server->name);
		start_server (server);
		break;
	    }
	    server = server->next;
	}
	free (msg[1]);
    }
}

/*
 * Signal-handler for SIGCHLD
 */

static void
sigchild (int foo)
{
    bosdebug ("child died\n");

    dead_children++;

    LWP_NoYieldSignal ((char *) theriper);
}

/*
 * bootstrap bosserver
 * Create riper thread and start all servers.
 */

static int
bosserver_init (void)
{
    struct mtype *server;
    int i;

    i = LWP_CreateProcess (theriper, 0, 1, NULL, "the-riper", &riperprocess);
    if (i)
	errx (1, "LWP_CreateProcess returned %d", i);

    server = servers;
    while (server) {
	i = start_server(server);
	if (i)
	    bosdebug ("server_server: returned %s (%d)", strerror (i), i);
	server = server->next;
    }
    
    return 0;
}


/*
 *
 */

static struct rx_service *bosservice = NULL;

static char *cell = NULL;
static char *realm = NULL;
static int do_help = 0;
static char *srvtab_file = NULL;
static int no_auth = 0;

static struct getargs args[] = {
    {"cell",	0, arg_string,    &cell, "what cell to use", NULL},
    {"realm",  0, arg_string,    &realm, "what realm to use"},
    {"prefix",'p', arg_string,    &bosserverprefix, 
     "directory where servers is stored", NULL},
    {"noauth",   0,  arg_flag,	  &no_auth, "disable authentication checks"},
    {"debug", 'd', arg_flag,      &debug, "output debugging"},
    {"help",  'h', arg_flag,      &do_help, "help"},
    {"srvtab",'s', arg_string,    &srvtab_file, "what srvtab to use"},
    { NULL, 0, arg_end, NULL }
};

static void
usage(int exit_code)
{
    arg_printusage(args, NULL, "", ARG_AFSSTYLE);
    exit (exit_code);
}

/*
 *
 */

int
main (int argc, char **argv)
{
    int ret;
    int optind = 0;
    
    if (getarg (args, argc, argv, &optind, ARG_AFSSTYLE)) {
	usage (1);
    }

    argc -= optind;
    argv += optind;

    if (argc) {
	printf("unknown option %s\n", *argv);
	return 1;
    }
    
    if (do_help)
	usage(0);

    if (bosserverprefix == NULL)
	bosserverprefix = MILKO_LIBEXECDIR;

    read_config_file ("/usr/arla/etc/bos.conf");

    ports_init();
    cell_init(0);

    if (no_auth)
	sec_disable_superuser_check ();

    if (cell)
	cell_setthiscell (cell);

    network_kerberos_init (srvtab_file);

    signal (SIGCHLD, sigchild);

    ret = network_init(htons(afsbosport), "bos", BOS_SERVICE_ID, 
		       BOZO_ExecuteRequest, &bosservice, realm);
    if (ret)
	errx (1, "network_init failed with %d", ret);
    
    ret = bosserver_init();
    if (ret)
	errx (1, "bosserver_init: error %d\n", ret);

    printf("Milko bosserver %s-%s started\n", PACKAGE, VERSION);

    rx_SetMaxProcs(bosservice,5) ;
    rx_StartServer(1) ;

    abort() ; /* should not get here */

    return 0;
}

