/*
 * Copyright (c) 2000 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "fsrv_locl.h"

RCSID("$Id: salvage.c,v 1.4 2000/03/08 23:10:16 lha Exp $");

struct inodeinfo {
    struct inodeinfo *parentp, *childp, *siblingp;
    struct inodeinfo *lfnodes;
    int32_t inode_num;
    int32_t parent_num;
    int32_t parent_unique;
    int32_t filetype;
    enum { IUNUSED = 0, IFOUND, ITREE } status;
};

static struct inodeinfo *
find_node (struct inodeinfo *inodes, int32_t inode, u_int32_t numnodes)
{
    u_int32_t i;

    for (i = 0; i < numnodes; i++) {
	if (inodes[i].inode_num == inode)
	    return &inodes[i];
    }
    return NULL;
}

#define CHECK_VALUE_EQ(check_val,val,string,modified)			\
do {									\
    if ((val) != (check_val)) {						\
        (check_val) = (val);						\
	modified = 1;							\
	mlog_log (MDEBSALVAGE, "%s had a errorous value resetting");	\
    }									\
} while (0)

#define CHECK_VALUE_BITMASK(check_val,val,string,modified)		\
do {									\
    if (((check_val) & ~(val)) != 0) {					\
	(check_val) &= (val);						\
	modified = 1;							\
	mlog_log (MDEBSALVAGE, "%s had a errorous value resetting");	\
    }									\
} while (0)

#define SET_VALUE_BITMASK(check_val,val,defval,string,modified)		\
do {									\
    if (((check_val) & (val)) == 0) {					\
	(check_val) |= (defval);					\
	modified = 1;							\
	mlog_log (MDEBSALVAGE, "%s had a missing value setting");	\
    }									\
} while (0)


static int
read_nodes_dir (u_int32_t num, struct voldb_entry *entry,
		struct inodeinfo *freenode)
{
    int mod = 0;

    assert (entry->type == TYPE_DIR);

    if (entry->u.file.FileType == 0) {
	mlog_log (MDEBSALVAGE, "%d is a unused node (dir)", num);
	return 1;
    }


    CHECK_VALUE_EQ(entry->u.dir.InterfaceVersion,1, "InterfaceVersion",mod);
    CHECK_VALUE_BITMASK(entry->u.dir.UnixModeBits,047777, "Unix rights",mod);
    SET_VALUE_BITMASK(entry->u.dir.UnixModeBits,040000,040000,
		      "Directory bit",mod);
    
    CHECK_VALUE_EQ(entry->u.dir.FileType,TYPE_DIR,"FileType",mod);

    freenode->inode_num 	= num;
    freenode->parent_num 	= dir_afs2local (entry->u.dir.ParentVnode);
    freenode->parent_unique	= entry->u.dir.ParentUnique;
    freenode->status	= IFOUND;
    freenode->filetype	= TYPE_DIR;
    
    if (mod) {
	/* XXX store bits to disk */
    }
    
    return 0;
}

static int
read_nodes_file (u_int32_t num, struct voldb_entry *entry,
		 struct inodeinfo *freenode)
{
    int mod = 0;
    assert (entry->type == TYPE_FILE);

    if (entry->u.file.FileType == 0) {
	mlog_log (MDEBSALVAGE, "%d is a unused node (file)", num);
	return 1;
    }
    
    CHECK_VALUE_EQ(entry->u.file.InterfaceVersion,1, "InterfaceVersion",mod);
    CHECK_VALUE_BITMASK(entry->u.file.UnixModeBits,047777, "Unix rights",mod);
    SET_VALUE_BITMASK(entry->u.file.UnixModeBits,0120000,0100000,
		      "Directory bit",mod);
    
    switch (entry->u.file.FileType) {
    case TYPE_FILE:
    case TYPE_LINK:
	break;
    default:
	CHECK_VALUE_EQ(entry->u.file.FileType,TYPE_FILE,"FileType",mod);
    }

    freenode->inode_num 	= num;
    freenode->parent_num 	= file_afs2local (entry->u.file.ParentVnode);
    freenode->parent_unique	= entry->u.file.ParentUnique;
    freenode->status	= IFOUND;
    freenode->filetype	= entry->u.file.FileType;

    if (mod) {
	/* XXX store bits to disk */
    }

    return 0;
}

static int
read_nodes (struct voldb *db, u_int32_t size,
	    struct inodeinfo **freenodes, u_int32_t *foundnodes,
	    int (*func) (u_int32_t, struct voldb_entry *, struct inodeinfo *))
{
    struct inodeinfo *freenode = *freenodes;
    u_int32_t i;
    int ret;

    for (i = 0; i < size; i++) {
	struct voldb_entry entry;
	
	ret = voldb_get_entry (db, i, &entry);
	if (ret) {
	    mlog_warn (MDEBSALVAGE, ret, "tree_connectivity: "
		       "get_entry failed");
	    continue;
	}

	ret = (*func) (i, &entry, freenode);
	if (ret) {
	    /* XXX should we really do nothing here */
	} else {
	    (*foundnodes)++;
	    freenode++;
	}
    }
    *freenodes = freenode ;
    return 0;
}


static void
check_dir_func (VenusFid *fid, const char *name, void *arg)
{

}

/*
 *
 */

static int
check_dir (volume_handle *vol, struct inodeinfo *node)
{
    struct mnode *n;
    struct msec m;
    fbuf the_fbuf;
    VenusFid fid;
    int ret;

    fid.Cell = 0;
    fid.fid.Volume = vol->vol;
    fid.fid.Vnode = dir_local2afs(node->inode_num);
    fid.fid.Unique = 0;

    mlog_log (MDEBSALVAGE, "checking dir %d.%d", fid.fid.Volume, fid.fid.Vnode);

    memset (&n, 0, sizeof(n));

    mnode_find (&fid.fid, &n);

    m.flags = VOLOP_WRITE|VOLOP_GETSTATUS|VOLOP_NOCHECK|VOLOP_PARENT;

    ret = vld_open_vnode (vol, n, &m);
    if (ret) {
	mnode_free (n, FALSE);
	return 1;
    }

    ret = fbuf_create (&the_fbuf, n->fd, n->sb.st_size,
		       FBUF_READ|FBUF_WRITE|FBUF_SHARED);

    ret = fdir_readdir (&the_fbuf, check_dir_func, NULL, &fid);
    if (ret)
	mlog_log (MDEBSALVAGE, "check_dir: fbuf_readdir failed");

    /*
     * XXX mdir_readdir to make sure that the list node->childp and
     * the content of the directory is right
     *
     * XXX also remeber to check for .__afsNNN files since that is
     * silly rename files, and those should be nuked.
     */

    mnode_free (n, FALSE);
    return ret;
}

static int
check_file (volume_handle *vol, struct inodeinfo *node)
{
    /* XXX Make sure the node exists */

    return 0;
}

/*
 *
 */

int
tree_connectvity (struct volume_handle *volh)
{
    u_int32_t dsize, fsize, size;
    u_int32_t i;
    u_int32_t foundnodes = 0;
    int ret;
    struct inodeinfo *freenode, *inodes = NULL;
    struct inodeinfo *lfnodes = NULL;

    ret = vld_db_uptodate (volh);
    if (ret) {
	mlog_warnx (MDEBERROR, "tree_connectity: vld_db_uptodate"
		    "failed with %d on volume %d",
		    ret, VLD_VOLH_NUM(volh));
	return ret;
    }

    ret = voldb_header_info(VLD_VOLH_DIR(volh), &dsize, NULL);
    if (ret) {
	mlog_warnx (MDEBERROR, "tree_connectity: voldb_header_info "
		    "failed with %d on volume %d (dir)",
		    ret, VLD_VOLH_NUM(volh));
	return ret;
    }

    ret = voldb_header_info(VLD_VOLH_FILE(volh), &fsize, NULL);
    if (ret) {
	mlog_warnx (MDEBERROR, "tree_connectity: voldb_header_info "
		    "failed with %d on volume %d (file)",
		    ret, VLD_VOLH_NUM(volh));
	return ret;
    }

    size = dsize + fsize;

    freenode = inodes = calloc (size, sizeof(struct inodeinfo));
    if (inodes == NULL) {
	mlog_warn (MDEBERROR, errno, "tree_connectity: calloc failed "
		   "(on volume %d with size %d", VLD_VOLH_NUM(volh), size);
	return ENOMEM;
    }
    
    ret = read_nodes (VLD_VOLH_DIR(volh),
		      dsize,
		      &freenode,
		      &foundnodes,
		      read_nodes_dir);
    if (ret)
	abort(); /* XXX */
    ret = read_nodes (VLD_VOLH_FILE(volh),
		      fsize,
		      &freenode,
		      &foundnodes,
		      read_nodes_file);
    if (ret)
	abort(); /* XXX */

    mlog_log (MDEBSALVAGE, "tree_connectvity: status: found %d nodes",
	      foundnodes);

    for (i = 0; i < foundnodes; i++) {
	struct inodeinfo *p, *n = &inodes[i] ;

	/* Skip root node */
	if (i == 0) {
	    n->status = ITREE;
	    continue;
	}
	
	p = find_node (inodes, n->parent_num, foundnodes);
	if (p == NULL) {
	    inodes[i].lfnodes = lfnodes;
	    lfnodes = &inodes[i];
	} else {
	    n->status = ITREE;
	    n->siblingp = p->childp;
	    p->childp = n;
	    n->parentp = p;
	}
    }
    for (i = 0; i < foundnodes; i++) {
	switch (inodes[i].status) {
	case ITREE:

	    /* XXX Run over all dir nodes to make sure that the
	     * content is what expected */

	    if (inodes[i].filetype == TYPE_DIR)
		check_dir (volh, &inodes[i]);
	    else {
		check_file (volh, &inodes[i]);
	    }

	    break;
	case IFOUND:
	    /* 
	     * this is a lost node 
	     *  XXX attach it to the lost+found directory 
	     */

	    /* XXX make sure is on the lfnodes list */

	    mlog_log (MDEBVLD, "lost inode found");
	    break;
	case IUNUSED:
	    abort();
	    break;
	default:
	    abort();   
	}

    }
    free (inodes);
    return 0;
}

