/*
 * Copyright (c) 1999, 2000 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Callback module
 *
 * This code has a MAJOR problem, it only keeps a hashtable on the first
 * address on the client, this is fine with singled ip clients.
 * But with afs3.5 you have the possibility to have multiple ip-addresses
 * per client. And then you will have a performance hit.
 *
 * This module also has no way for the rx module to get out stuff like
 * netmasks and MTU, that should be very interesting to know performance-wise.
 */

#include <config.h>

#include <stdio.h>
#include <sys/types.h>
#include <assert.h>

#include <atypes.h>
#include <fs.h>
#include <cb.h>

#include <hash.h>
#include <heap.h>
#include <list.h>

#include <ports.h>
#include <service.h>

#include <rx/rx.h>
#include <rx/rx_null.h>
#include <rx/rxgencon.h>

#include <cb.cs.h>

#include <err.h>

#include <ropa.h>

RCSID("$Id: ropa.c,v 1.5 2000/03/20 22:23:17 lha Exp $");

/*
 * Declarations
 */

#define ROPA_STACKSIZE (16*1024)

struct ropa_client {
    interfaceAddr 	addr;		/* interfaces and uuid of client */
    u_int16_t 		port;		/* port of client */
    time_t		lastseen;	/* last got a message */
    List		*callbacks;	/* list of ccpairs */
    int			ref;		/* refence counter */
    Listitem	       	*li; 		/* where on lru */
};

struct ropa_ccpair {			/* heap object */
    struct ropa_client 	*client;	/* pointer to client */
    struct ropa_cb	*cb;		/* pointer to callback */
    time_t		expire;		/* when this cb expire */
    int32_t		version;	/* version of this callback */
    heap_ptr		heap;		/* heap pointer */
    Listitem	       	*li;		/* where on lru */
};

struct ropa_cb {
    AFSFid		fid;		/* what fid this callback is on */
    List 		*ccpairs;	/* list of clients holding this cb */
    int 		ref;		/* refence counter */
    Listitem	       	*li;		/* where on lru */
};

static void break_callback (struct ropa_cb *cb,struct ropa_client *caller);
static void break_ccpair (struct ropa_ccpair *cc, Bool notify_clientp);
static void break_client (struct ropa_client *c, Bool notify_clientp);
static void create_callbacks (void);
static void create_ccpairs (void);
static int uuid_magic_eq (afsUUID *uuid1, afsUUID *uuid2);


/*
 * Module variables
 */

#define NUM_CLIENTS	100
#define NUM_CALLBACKS	300
#define NUM_CCPAIR	600

/*
 *
 */

static Hashtab *ht_clients_ip = NULL;
static Hashtab *ht_clients_uuid = NULL;
static Hashtab *ht_callbacks = NULL;
/* last used on tail */
static List *lru_clients = NULL;
static List *lru_ccpair = NULL;
static List *lru_callback = NULL;
static unsigned long num_clients = 0;
static unsigned long num_callbacks = 0;
static unsigned long num_ccpair = 0;

static Heap *heap_ccpairs = NULL;
static PROCESS cleaner_pid;

/*
 *
 */

static int
uuid_magic_eq (afsUUID *uuid1, afsUUID *uuid2)
{
    return memcmp (&uuid1->node,
		   &uuid2->node,
		   sizeof (uuid1->node));
}

/*
 * help functions for the hashtab
 */

/*
 * We compare using addresses
 */

static int
clients_cmp_ip (void *p1, void *p2)
{
    struct ropa_client *c1 = (struct ropa_client *) p1;
    struct ropa_client *c2 = (struct ropa_client *) p2;
    int i,j;

    for (i = 0; i < c1->addr.numberOfInterfaces; i++) {
	for (j = 0; j < c2->addr.numberOfInterfaces; j++) {
	    if (c1->addr.addr_in[i] == c2->addr.addr_in[j]
		&& c1->port == c2->port) {
		return 1;
	    }
	}
    }
    return 0;
}

/*
 * Might hash on UUID instead
 */

static unsigned
clients_hash_ip (void *p)
{
    struct ropa_client *c = (struct ropa_client *)p;

    assert (c->addr.numberOfInterfaces);
    
    return c->addr.addr_in[0] + c->port;
}

/*
 * We compare using addresses
 */

static int
clients_cmp_uuid (void *p1, void *p2)
{
    return uuid_magic_eq (&((struct ropa_client *) p1)->addr.uuid,
			  &((struct ropa_client *) p2)->addr.uuid);
}

static unsigned
clients_hash_uuid (void *p)
{
    struct ropa_client *c =  (struct ropa_client *) p;

    return 
	c->addr.uuid.node[0] + 
	(c->addr.uuid.node[1] << 6) +
	(c->addr.uuid.node[2] << 12) +
	(c->addr.uuid.node[3] << 18) +
	(c->addr.uuid.node[4] << 24) +
	((c->addr.uuid.node[5] << 30) & 0x3);
}

/*
 * 
 */

static int
callbacks_cmp (void *p1, void *p2)
{
    struct ropa_cb *c1 = (struct ropa_cb *) p1;
    struct ropa_cb *c2 = (struct ropa_cb *) p2;

    return c1->fid.Volume - c2->fid.Volume &&
	c1->fid.Vnode - c2->fid.Vnode &&
	c1->fid.Unique - c2->fid.Unique;
}

/*
 *
 */

static unsigned
callbacks_hash (void *p)
{
    struct ropa_cb *c = (struct ropa_cb *)p;
    
    return c->fid.Volume + c->fid.Vnode +
	c->fid.Unique;
}

/*
 *
 */

static int
ccpair_cmp (const void *p1, const void *p2)
{
    struct ropa_ccpair *c1 = (struct ropa_ccpair *) p1;
    struct ropa_ccpair *c2 = (struct ropa_ccpair *) p2;

    return c1->expire - c2->expire;
}

/*
 *
 */

static Bool
client_inuse_p (struct ropa_client *c)
{
    assert (c);
    return c->port == 0 ? FALSE : TRUE ;
}

/*
 *
 */

static Bool
callback_inuse_p (struct ropa_cb *cb)
{
    assert (cb);
    return !listemptyp (cb->ccpairs) ;
}

/*
 *
 */

static Bool
ccpairs_inuse_p (struct ropa_ccpair *cc)
{
    assert (cc);
    return cc->client == NULL ? FALSE : TRUE ;
}

/*
 *
 */

static void
create_clients (void)
{
    struct ropa_client *c;
    unsigned long i;

    c = calloc (NUM_CLIENTS, sizeof (*c));
    if (c == NULL)
	err (1, "create_clients: calloc");

    for (i = 0 ; i < NUM_CLIENTS; i++) {
	memset (&c->addr, 0, sizeof(c->addr));
	c[i].port = 0;
	c[i].lastseen = 0;
	c[i].callbacks = listnew();
	if (c[i].callbacks == NULL)
	    err (1, "create_clients: listnew");
	c[i].ref = 0;
	c[i].li = listaddtail (lru_clients, &c[i]);
    }
    num_clients += NUM_CLIENTS;
}

/*
 *
 */

static void
create_callbacks (void)
{
    struct ropa_cb *c;
    int i;

    c = calloc (NUM_CALLBACKS, sizeof (*c));
    if (c == NULL)
	err (1, "create_callbacks: calloc");

    for (i = 0; i < NUM_CALLBACKS ; i++) {
	memset (&c[i], 0, sizeof (c[0]));
	c[i].ccpairs = listnew();
	if (c[i].ccpairs == NULL)
	    err (1, "create_callbacks: listnew");
	c[i].li = listaddtail (lru_callback, &c[i]);
    }
    num_callbacks += NUM_CALLBACKS;
}

/*
 *
 */

void
create_ccpairs (void) 
{
    struct ropa_ccpair *c;
    int i;

    c = calloc (NUM_CCPAIR, sizeof (*c));
    if (c == NULL)
	err (1, "create_ccpairs: calloc");

    for (i = 0; i < NUM_CCPAIR ; i++) {
	memset (&c[i], 0, sizeof (c[0]));
	c[i].li = listaddtail (lru_ccpair, &c[i]);
    }
    num_ccpair += NUM_CCPAIR;
}

/*
 *
 */

static void
client_ref (struct ropa_client *c)
{
    assert (c->ref >= 0);
    assert (c->li != NULL || c->ref == 0);
    if (c->li)
	listdel (lru_clients, c->li);
    c->ref++;
    if (c->li)
	c->li = listaddhead (lru_clients, c);
}

/*
 *
 */

static void
callback_ref (struct ropa_cb *c)
{
    assert (c->ref >= 0);
    assert (c->li != NULL || c->ref == 0);
    if (c->li)
	listdel (lru_callback, c->li);
    c->ref++;
    if (c->li)
	c->li = listaddhead (lru_callback, c);
}

/*
 *
 */

static void
client_deref (struct ropa_client *c)
{
    c->ref--;

    if (c->ref == 0) {
	int ret = hashtabdel (ht_clients_ip, c);
	assert (ret == 0);
	ret = hashtabdel (ht_clients_uuid, c);
	assert (ret == 0);
	listdel (lru_clients, c->li);
	c->port = 0;
	c->li = listaddtail (lru_clients, c);
    }
}

/*
 *
 */

static void
callback_deref (struct ropa_cb *cb)
{
    cb->ref--;

    if (cb->ref == 0) {
	int ret = hashtabdel (ht_callbacks, cb);
	assert (ret == 0);
	assert (listemptyp (cb->ccpairs));
	cb->li = listaddtail (lru_callback, cb);
    }
}

/*
 * 
 */

static struct ropa_ccpair *
add_client (struct ropa_cb *cb, struct ropa_client *c)
{
    struct timeval tv;
    struct ropa_ccpair *cc;

    assert (cb && c);

    callback_ref (cb);
    client_ref (c);

    cc = listdeltail (lru_ccpair);
    assert (cc);

    if (ccpairs_inuse_p (cc))
	break_ccpair (cc, TRUE);

    /* XXX  do it for real */
    gettimeofday(&tv, NULL);
    cc->expire = tv.tv_sec + 3600;
    cc->version += 1;
    
    heap_insert (heap_ccpairs, cc, &cc->heap);
    LWP_NoYieldSignal (heap_ccpairs);
    listaddtail (cb->ccpairs, cc);
    
    cc->client = c;
    cc->cb = cb;
    cc->li = listaddhead (lru_ccpair, cc);

    return cc;
}

/*
 * performance sucks, but it works
 */

int
ropa_getcallback (u_int32_t host, u_int16_t port, const struct AFSFid *fid,
		  AFSCallBack *callback)
{
    struct ropa_client ckey, *c;
    struct ropa_cb cbkey, *cb;
    struct ropa_ccpair *cc ;
    int ret;

    ckey.addr.numberOfInterfaces = 1;
    ckey.addr.addr_in[0] = host;
    ckey.port = port;

    c = hashtabsearch (ht_clients_ip, &ckey);
    if (c == NULL) {
	interfaceAddr remote;
	struct rx_connection *conn;
	
	conn = rx_NewConnection (host, port, CM_SERVICE_ID,
				 rxnull_NewClientSecurityObject(),
				 0);
	
	ret = RXAFSCB_WhoAreYou (conn, &remote);
    
	if (ret == RX_CALL_DEAD) {
	    rx_DestroyConnection (conn);
	    return ENETDOWN;
	} else if (ret == RXGEN_OPCODE) {
	    ret = RXAFSCB_InitCallBackState(conn);
	    rx_DestroyConnection (conn);
	    if (ret)
		return ret;

	    remote.numberOfInterfaces = 1;
	    remote.addr_in[0] = host;
	    remote.subnetmask[0] = 0xffffff00;
	    remote.mtu[0] = 1500;
	    remote.uuid.node[0] = 0xff & host;
	    remote.uuid.node[1] = 0xff & (host >> 8);
	    remote.uuid.node[2] = 0xff & (host >> 8);
	    remote.uuid.node[3] = 0xff & (host >> 8);
	    remote.uuid.node[4] = 0xaa;
	    remote.uuid.node[5] = 0x77;

	    ckey.addr = remote;

	    c = hashtabsearch (ht_clients_uuid, &ckey);
	    if (c == NULL) {
		c = listdeltail (lru_clients);
		if (c == NULL)
		    abort();
		c->li = NULL;
		if (client_inuse_p (c))
		    break_client (c, TRUE);
		c->addr = remote;
		c->port = port;
		hashtabadd (ht_clients_ip, c);
		hashtabadd (ht_clients_uuid, c);
		c->li = listaddhead (lru_clients, c);
	    } else {
		listdel (lru_clients, c->li);
		c->li = listaddhead (lru_clients, c);
	    }	    
	} else {
	    ckey.addr = remote;
	    c = hashtabsearch (ht_clients_uuid, &ckey);
	    if (c != NULL) {
		if (uuid_magic_eq(&remote.uuid, &c->addr.uuid)) {
		    struct ropa_client d;

		    d.addr = remote;
		    d.port = port;
		    
		    /* XXX the lad has a new interface, just add it */
		    if (clients_cmp_ip (c, &d) == FALSE) {
			c->addr.addr_in[c->addr.numberOfInterfaces] = host;
			c->addr.numberOfInterfaces += 1;
		    }
		    rx_DestroyConnection (conn);
		} else {
		    ret = RXAFSCB_InitCallBackState3 (conn, &remote.uuid);
		    rx_DestroyConnection (conn);
		    switch (ret) {
		    case 0:
			break;
		    case RX_CALL_DEAD:
			return ENETDOWN;
		    default:
			return ENETDOWN;
		    }
		    break_client (c, FALSE);
		    c = listdeltail (lru_clients);
		    if (client_inuse_p (c))
			break_client (c, TRUE);
		    c->addr = remote;
		    hashtabadd (ht_clients_ip, c);
		    hashtabadd (ht_clients_uuid, c);
		    c->li = listaddhead (lru_clients, c);
		}
	    } else {
		c = listdeltail (lru_clients);
		if (client_inuse_p (c))
		    break_client (c, TRUE);
		c->li = listaddhead (lru_clients, c);
	    }
	}
    }

    cbkey.fid = *fid;

    cb = hashtabsearch (ht_callbacks, &cbkey);
    if (cb == NULL) {
	cb = listdeltail (lru_callback);
	if (cb == NULL)
	    abort();
	cb->li = NULL;
	if (callback_inuse_p (cb))
	    break_callback (cb, NULL);
	callback_ref(cb);
	hashtabadd (ht_callbacks, cb);
	cb->li = listaddhead (lru_callback, cb);
    } else {
	callback_ref(cb);
    }

    cc = add_client (cb, c);
    if (cc == NULL)
	abort();

    callback_deref (cb);

    callback->CallBackVersion = cc->version;
    callback->ExpirationTime = cc->expire;
    callback->CallBackType = CBSHARED;
    return 0;
}

/*
 * Notify client
 */

static int
notify_client (struct ropa_client *c, AFSCBFids *fids, AFSCBs *cbs)
{
    struct rx_connection *conn;
    int16_t port = c->port;
    int i, ret;

    if (port == afsport)
	port = afscallbackport;

    for (i = 0; i < c->addr.numberOfInterfaces ; i++) {
	conn = rx_NewConnection (c->addr.addr_in[i],
				 htons (port),
				 CM_SERVICE_ID,
				 rxnull_NewClientSecurityObject(),
				 0);

	ret = RXAFSCB_CallBack (conn, fids, cbs);
	rx_DestroyConnection (conn);
	if (ret == 0) 
	    break;

	/* XXX warn */
    }

    return ret;
}

/*
 *
 */

static void
break_ccpair (struct ropa_ccpair *cc, Bool notify_clientp)
{
    AFSCBFids fids;
    AFSCBs cbs;

    cc->expire = 0;

    listdel (lru_ccpair, cc->li);

    if (notify_clientp) {
	AFSCallBack callback;
	callback.CallBackVersion = cc->version;
	callback.ExpirationTime  = cc->expire;
	callback.CallBackType    = CBDROPPED;

	fids.len = 1;
	fids.val = &cc->cb->fid;
	cbs.len = 1;
	cbs.val = &callback;
	notify_client (cc->client, &fids, &cbs);
    }

    cc->client = NULL;
    cc->cb = NULL;

    client_deref (cc->client);
    cc->client = NULL;
    callback_deref (cc->cb);
    cc->cb = NULL;
    heap_remove (heap_ccpairs, cc->heap);
    listaddtail (lru_ccpair, cc);
}

static int
add_to_cb (AFSFid *fid, AFSCallBack *cb, AFSCBFids *fids, AFSCBs *cbs)
{
    AFSFid      *n_fid;
    AFSCallBack *n_cb;

    n_fid = realloc (fids->val, fids->len + 1);
    if (n_fid == NULL)
	return ENOMEM;
    fids->val = n_fid;
    n_cb  = realloc (cbs->val, cbs->len + 1);
    if (n_cb == NULL)
	return ENOMEM;
    cbs->val = n_cb;
    fids->val[fids->len] = *fid;
    cbs->val[cbs->len] = *cb;
    ++fids->len;
    ++cbs->len;
    return 0;
}

static void
break_ccpairs (struct ropa_client *c, List *ccpairs, Bool notify_clientp)
{
    AFSCBFids fids;
    AFSCBs    cbs;
    struct ropa_ccpair *cc;
    AFSCallBack callback;

    fids.val = NULL;
    fids.len = 0;

    cbs.val = NULL;
    cbs.len = 0;

    callback.CallBackType = CBDROPPED;
    while ((cc = listdeltail (c->callbacks)) != NULL) {
	add_to_cb (&cc->cb->fid, &callback, &fids, &cbs);
	break_ccpair (cc, FALSE);
    }
    
    if (notify_clientp)
	notify_client (c, &fids, &cbs);
}


/*
 * break the callback and remove it from the chain of clients.
 * it the clients responsible for the callback is the caller
 * don't break the callback.
 */

static void
break_callback (struct ropa_cb *cb, struct ropa_client *caller)
{
    struct ropa_ccpair *cc;

    assert (cb);
    assert (cb->ccpairs);

    callback_ref (cb);
    client_ref (caller);

    cb = cc->cb;

    while ((cc = listdelhead (cb->ccpairs)) != 0)
	break_ccpair (cc, cc->client != caller);

    callback_deref (cb);
    client_deref (caller);
}

/*
 *
 */

void
ropa_break_callback (AFSFid *fid, u_int32_t addr, u_int16_t port)
{
    struct ropa_client ckey, *c;
    struct ropa_cb cbkey, *cb;

    ckey.addr.numberOfInterfaces = 1;
    ckey.addr.addr_in[0] = addr;
    ckey.port = port;

    c = hashtabsearch (ht_clients_ip, &ckey);
    if (c == NULL) {
	/* XXX warn */
	return;
    }

    cbkey.fid = *fid;
    
    cb = hashtabsearch (ht_clients_ip, &cbkey);
    if (cb == NULL) {
	/* XXX warn */
	return;
    }

    break_callback (cb, c);
}

/*
 *
 */

static void
break_client (struct ropa_client *c, Bool notify_clientp)
{
    assert (c);
    
    client_ref (c);

    break_ccpairs (c, c->callbacks, notify_clientp);

    c->port = 0;
    client_deref (c);
}

void
ropa_break_client (u_int32_t addr, u_int16_t port)
{
    struct ropa_client ckey, *c;
    
    ckey.addr.numberOfInterfaces = 1;
    ckey.addr.addr_in[0] = addr;
    ckey.port = port;
    
    c = hashtabsearch (ht_clients_ip, &ckey);
    if (c == NULL) {
	/* XXX warn */
	return;
    }

    break_client (c, TRUE);
}

/*
 *
 */

static void
heapcleaner (char *arg)
{
    const void *head;
    struct timeval tv;

    while (1) {

	while ((head = heap_head (heap_ccpairs)) == NULL)
	    LWP_WaitProcess (heap_ccpairs);
	
	while ((head = heap_head (heap_ccpairs)) != NULL) {
	    struct ropa_ccpair *cc = (struct ropa_ccpair *)head;
	    
	    gettimeofday (&tv, NULL);
	    
	    if (tv.tv_sec < cc->expire) {
		unsigned long t = cc->expire - tv.tv_sec;
		IOMGR_Sleep (t);
		break;
	    }

	    heap_remove_head (heap_ccpairs);
	    cc->heap = -1;
	}
    }
}


/*
 * init the ropa-module
 */

int
ropa_init (unsigned long num_callback, unsigned long num_clients)
{
    ht_callbacks = hashtabnew (num_callback, callbacks_cmp, callbacks_hash);
    if (ht_callbacks == NULL)
	errx (1, "ropa_init: failed to create hashtable for callbacks");

    ht_clients_ip = hashtabnew (num_clients, clients_cmp_ip, clients_hash_ip);
    if (ht_clients_ip == NULL)
	errx (1, "ropa_init: failed to create hashtable for clients_ip");
	
    ht_clients_uuid = hashtabnew (num_clients, clients_cmp_uuid, clients_hash_uuid);
    if (ht_clients_uuid == NULL)
	errx (1, "ropa_init: failed to create hashtable for clients_uuid");
	
    lru_clients = listnew ();
    if (lru_clients == NULL)
	errx (1, "ropa_init: failed to create lru for clients");
	
    lru_ccpair = listnew ();
    if (lru_ccpair == NULL)
	errx (1, "ropa_init: failed to create list for ccpairs");

    lru_callback = listnew ();
    if (lru_ccpair == NULL)
	errx (1, "ropa_init: failed to create list for ccpairs");

    heap_ccpairs = heap_new (NUM_CCPAIR, ccpair_cmp);
    if (heap_ccpairs == NULL)
	errx (1, "ropa_init: failed to create heap for ccpairs");

    create_clients();
    create_callbacks();
    create_ccpairs();

    if (LWP_CreateProcess (heapcleaner, ROPA_STACKSIZE, 1,
			   NULL, "heap-invalidator", &cleaner_pid))
	errx (1, "ropa_init: LWP_CreateProcess failed");


    return 0;
}

